#include "mm_intros_fun.h"

static vcpu_guest_context_t eip_ctxt[10];
static unsigned long cr3 = -1;

void *mapmem_domu(
    int xc_handle,
    int cpu,
    int current_domid,
    void *guest_va,
    int perm,
    pgd_access pgd)
{
  //unsigned long va = (unsigned long) guest_va;
    //long npgs = xc_get_tot_pages(xc_handle, current_domid);

    //Assume Paging enabled. Check paging mode.
    if ( mode == MODE_UNKNOWN )
    {
        xen_capabilities_info_t caps;
        (void)xc_version(xc_handle, XENVER_capabilities, caps);

        if ( strstr(caps, "-x86_32p") ) {
            mode = MODE_PAE;
	}
        else if ( strstr(caps, "-x86_32") ) {
            mode = MODE_32;
	}
    }

    if ( mode == MODE_PAE )
        return mapmem_domu_pae(xc_handle, cpu, guest_va, perm, current_domid, pgd);

    return mapmem_domu_rp(xc_handle, cpu, guest_va, perm, current_domid, pgd);
}

static void *mapmem_domu_rp(
	int xc_handle,
    	int cpu,
    	void *guest_va,
    	int perm, 
	int current_domid,
    	pgd_access pgd)
{
    unsigned long l2e, l1e, va = (unsigned long)guest_va;
    uint32_t *l2, *l1;
    static void *v[MAX_VIRT_CPUS];
    vcpu_guest_context_t ctxt[10];
 
    if (xc_vcpu_getcontext(xc_handle, current_domid, cpu, ctxt) != 0) {
       	printf("ERROR: Failed to retrieve vcpu context.\n");
	return NULL;
    }

    if(pgd == PROCESS_PGD) {
	cr3 = take_mfnpfn_from_address(ctxt[cpu].ctrlreg[3]); 	//Retrieve a correct MFN for a process address.
								//CR3 taken from process vcpu_context is a machine address.
    }

    else { 
	if(cr3 == -1) {
		if(get_mfn_cr3_kernel(xc_handle, current_domid, perm, &cr3) == -1) {
			return NULL;
    		}
    	}
    }

    /*Map pgd in your space address from cr3 domU context.*/
    l2 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, cr3);
    if ( l2 == NULL ) {
	printf("l2 NULL Exit.\n");
        return NULL;
    }
 
    /*
      Generate pgd entry address:
	1. Shift >> 22 for the virtual address: take first 10bits used as displacement in pgd table.
	2. Add offset calculated above to cr3 address (base) and then have pgd entry address: use array notation to access memory.
    */
    l2e = l2[l2_table_offset_i386(va)];
    munmap(l2, PAGE_SIZE);
    if ( !(l2e & _PAGE_PRESENT) ) {
	printf("ERROR: l2e NOT PRESENT\n");
        return NULL;
    }

    //Map PGD entry address in your address space.
    l1 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, l2e >> PAGE_SHIFT);
    if ( l1 == NULL ) {
	printf("l1 NULL Exit.\n");
        return NULL;
    }
 
    //As described above for level 1.
    l1e = l1[l1_table_offset_i386(va)];
    munmap(l1, PAGE_SIZE);
    if ( !(l1e & _PAGE_PRESENT) ) {
	printf("ERROR: l1e NOT PRESENT\n");;
        return NULL;
    }

    if ( v[cpu] != NULL )
        munmap(v[cpu], PAGE_SIZE);
    v[cpu] = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, perm, l1e >> PAGE_SHIFT);
    if ( v[cpu] == NULL ) {
	printf("Page Table NULL Exit.\n");
        return NULL;
    }

    return (void *)((unsigned long)v[cpu] | (va & (PAGE_SIZE - 1)));
}

static void *mapmem_domu_pae(
    int xc_handle,
    int cpu,
    void *guest_va,
    int perm,
    int current_domid,
    pgd_access pgd)
{
    uint64_t l3e, l2e, l1e;
    unsigned long va = (unsigned long)guest_va;
    uint64_t *l3, *l2, *l1;
    static void *v[MAX_VIRT_CPUS];
    vcpu_guest_context_t ctxt[10];

    if (xc_vcpu_getcontext(xc_handle, current_domid, cpu, ctxt) != 0) {
       	printf("ERROR: Failed to retreive vcpu context.\n");
	return NULL;
    }

    if(pgd == PROCESS_PGD) {
	cr3 = take_mfnpfn_from_address(ctxt[cpu].ctrlreg[3]); 	//Retrieve a correct MFN for a process address.
								//CR3 taken from process vcpu_context is a machine address.
    }

    else { 
	if(cr3 == -1) {
		if(get_mfn_cr3_kernel(xc_handle, current_domid, perm, &cr3) == -1) {
			return NULL;
		}
    	}
    }

    //Walking level is as described above for rp.
    l3 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, cr3);
    if ( l3 == NULL ) {
	printf("l3 NULL Exit.\n");
        return NULL;
    }

    l3e = l3[l3_table_offset_pae(va)];
    munmap(l3, PAGE_SIZE);
    if ( !(l3e & _PAGE_PRESENT) ) {
	printf("ERROR: l3e NOT PRESENT\n");
        return NULL;
    }

    l2 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, l3e >> PAGE_SHIFT);
    if ( l2 == NULL ) {
        printf("l2 NULL Exit.\n");
	return NULL;
    }

    l2e = l2[l2_table_offset_pae(va)];
    munmap(l2, PAGE_SIZE);
    if ( !(l2e & _PAGE_PRESENT) ) {
	printf("ERROR: l2e NOT PRESENT\n");
	return NULL;
    }

    l1 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, l2e >> PAGE_SHIFT);
    if ( l1 == NULL ) {
        printf("l1 NULL Exit.\n");
	return NULL;
    }

    l1e = l1[l1_table_offset_pae(va)];
    munmap(l1, PAGE_SIZE);
    if ( !(l1e & _PAGE_PRESENT) ) {
        return NULL;
	printf("ERROR: l1e NOT PRESENT\n");
    }

    if ( v[cpu] != NULL )
        munmap(v[cpu], PAGE_SIZE);

    v[cpu] = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, perm, l1e >> PAGE_SHIFT);
    if ( v[cpu] == NULL ) {
	printf("Page Table NULL Exit.\n");
        return NULL;
   }

    return (void *)((unsigned long)v[cpu] | (va & (PAGE_SIZE - 1)));
}

unsigned long convert_pfn_to_mfn(
	int xc_handle,
	int current_domid,
	int perm,
	unsigned long pfn)
{
    xc_dominfo_t info;
    shared_info_t *live_shinfo = NULL;
    unsigned long *live_pfn_to_mfn_frame_list_list = NULL;
    unsigned long *live_pfn_to_mfn_frame_list = NULL;

    /* Live mapping of the table mapping each PFN to its current MFN. */
    unsigned long *live_pfn_to_mfn_table = NULL;
    unsigned long nr_pfns = 0;
    unsigned long ret = -1;

    //Initialise domain info struct.
     if (xc_domain_getinfo(xc_handle, current_domid, 1, &info) != 1) {
        printf("ERROR: Failed to get domain info.\n");
        ret = -1;
        goto error_exit;
    }

    nr_pfns = info.max_memkb >> (XC_PAGE_SHIFT - 10);

    live_shinfo = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, info.shared_info_frame);

    if (live_shinfo == NULL){
        printf("ERROR: failed to init live_shinfo\n");
        goto error_exit;
    }
    //live_shinfo struct is defined in xen.h.
    live_pfn_to_mfn_frame_list_list = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, live_shinfo->arch.pfn_to_mfn_frame_list_list);
    if (live_pfn_to_mfn_frame_list_list == NULL) {
        printf("ERROR: failed to init live_pfn_to_mfn_frame_list_list\n");
        goto error_exit;
    }

    live_pfn_to_mfn_frame_list = xc_map_foreign_batch(xc_handle, current_domid, PROT_READ, live_pfn_to_mfn_frame_list_list, (nr_pfns+(PFN_FRAME*PFN_FRAME)-1)/(PFN_FRAME*PFN_FRAME));
    if (live_pfn_to_mfn_frame_list == NULL) {
        printf("ERROR: failed to init live_pfn_to_mfn_frame_list\n");
        goto error_exit;
    }

    live_pfn_to_mfn_table = xc_map_foreign_batch(xc_handle, current_domid, PROT_READ, live_pfn_to_mfn_frame_list, (nr_pfns+PFN_FRAME-1)/PFN_FRAME);
    if (live_pfn_to_mfn_table  == NULL) {
        printf("ERROR: failed to init live_pfn_to_mfn_table\n");
        goto error_exit;
    }

    ret = live_pfn_to_mfn_table[pfn];

error_exit:
    if (live_shinfo) munmap(live_shinfo, XC_PAGE_SIZE);
    if (live_pfn_to_mfn_frame_list_list)
        munmap(live_pfn_to_mfn_frame_list_list, XC_PAGE_SIZE);
    if (live_pfn_to_mfn_frame_list)
        munmap(live_pfn_to_mfn_frame_list, XC_PAGE_SIZE);

    return ret;
}

int get_mfn_cr3_kernel(
	int xc_handle,
	int current_domid,
	int perm,
	unsigned long *cr3)
{
    	//Take from System.map entry for XEN-Kernel 3.1: c042f0c4 B swapper_pg_dir.
    	//This is a kernel virtual address statically mapped, is a pseudo-physical address.
	//We need to convert PFN to MFN accessing to physical-to-machine table.
	/* 
	swapper_pg_dir symbol is a kernel virtual address statically mapped for PDPT but for kernel running on bare metal.
	For a kernel's domU this address is relocated in another virtual location for the kernel address space.
	To retrieve the virtual address relocated we need:
		1. Obtain PFN swapper_pg_dir and take offset.
		2. Convert PFN to MFN and map this MFN in my process space address.
		3. Add offset to MFN calculated above and access this location that contain a relocated pseudo-physical address for PDPT table.
		4. Take cr3 address base, normalize and calculate PFN for a real domU PDPT address.
		5. Convert PFN to MFN. This function don't map this MFN in process space address. Is mapped in mapmem_domu_X as first PT level.
	*/
	//[NOTE] RETURN A MFN NOT MAPPED!! In walk PT code this is mapped at first level access.
	unsigned long ksysmap_cr3, pfn_kcr3, mfn_kcr3, mfn_kcr3_mapped1, kpgd_addr_addr, cr3_phys_dom_reloc;
	unsigned long *mfn_kcr3_mapped;
	uint32_t offset;

	if((ksysmap_cr3 = get_kernel_address_from_symbol(map_file, "swapper_pg_dir")) == -1) {
		return -1;
	}

	ksysmap_cr3 = normalize_phys_address(ksysmap_cr3);

	//1.
	pfn_kcr3 = take_mfnpfn_from_address(ksysmap_cr3);
	offset = take_offset_from_address(ksysmap_cr3);

	//2.
	if((mfn_kcr3 = convert_pfn_to_mfn(xc_handle, current_domid, perm, pfn_kcr3)) == -1) {
		printf("ERROR: convert pfn to mfn! 1");
		return -1;
	}
	mfn_kcr3_mapped = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, mfn_kcr3);
	//printf("MFN MAPPED: %x\n", mfn_mapped);

	//3.
	mfn_kcr3_mapped1 = (unsigned long)mfn_kcr3_mapped;
	kpgd_addr_addr = mfn_kcr3_mapped1 + offset;
	//printf("INDIRIZZO CHE CONTIENE INDIRIZZO MACHINE KPGD: %x\n", kpgd_addr_addr);
	cr3_phys_dom_reloc = *((unsigned long*)kpgd_addr_addr);

	//4.
	//cr3_phys_dom_reloc = cr3_phys_dom_reloc & 0xFFFFFFE0;
	//printf("Indirizzo fisico rilocato per KPGD DomU: %x\n", cr3_phys_dom_reloc);
	cr3_phys_dom_reloc = normalize_phys_address(cr3_phys_dom_reloc);
	cr3_phys_dom_reloc = take_mfnpfn_from_address(cr3_phys_dom_reloc);

	//5.
	if((*cr3 = convert_pfn_to_mfn(xc_handle, current_domid, perm, cr3_phys_dom_reloc)) == -1) {
		printf("ERROR: convert pfn to mfn! 2");
		return -1;
	}

	return 0; //All ok.
}

unsigned long get_kernel_address_from_symbol(
	char *symbol_t,
	char *symbol)
{
	FILE *f_pointer;
	char kst_symbol_line[MAX_LINE_KT], kst_symbol_name[MAX_SYMBOL_KNAME], kst_symbol_address[MAX_SYMBOL_ADDRESS], kst_symbol_address_hex[MAX_SYMBOL_ADDRESS+2];
	int lenght;

	if((f_pointer = fopen(symbol_t, "r")) == NULL) {
		printf("ERROR: opening System Map Table: %s", symbol_t);
		return -1;
	}

	while(fgets(kst_symbol_line, MAX_LINE_KT, f_pointer) != NULL) {
		//Address and identifier is 11 position before kernel symbol name.
		//Copy the Symbol Name ('lenght-12') from position 11 of kst_symbol_line.
		lenght = strlen(kst_symbol_line);
		strncpy(kst_symbol_name, kst_symbol_line + 11, lenght - 12);
		kst_symbol_name[lenght-12] = '\0'; //Append NULL to the string.

		if(!strcmp(kst_symbol_name, symbol)) {
			strncpy(kst_symbol_address, kst_symbol_line, MAX_SYMBOL_ADDRESS);
			kst_symbol_address[MAX_SYMBOL_ADDRESS] = '\0';

			strncpy(kst_symbol_address_hex, "0x", 2); //Generate an hex format address.
			strncpy(kst_symbol_address_hex+2, kst_symbol_address, MAX_SYMBOL_ADDRESS);
			kst_symbol_address_hex[MAX_SYMBOL_ADDRESS+2] = '\0';

			return strtoul(kst_symbol_address_hex, NULL, 16); //Convert a string in base 16.
		}
	}

	printf("ERROR: Symbol %s not found!\n", symbol);
	return -1;
}

uint32_t *read_eip(
	int xc_handle,
	uint32_t cpu,
	uint32_t current_domid)
{
    if (xc_vcpu_getcontext(xc_handle, current_domid, cpu, eip_ctxt) != 0) {
       	printf("ERROR: Failed to retreive vcpu context.\n");
	return NULL;
    }

    return (uint32_t *)eip_ctxt[cpu].user_regs.eip;
}

int write_eip(
	int xc_handle,
	uint32_t cpu,
	uint32_t current_domid,
	uint32_t new_eip)
{
    eip_ctxt[cpu].user_regs.eip = new_eip;

    if (xc_vcpu_setcontext(xc_handle, current_domid, cpu, eip_ctxt) != 0) {
       	printf("ERROR: Failed to set vcpu context.\n");
	return -1;
    }

    return 0;
}

int is_kernel_address(unsigned long address) {
	if(address >= 0xc0000000) {
		return 1;
	}

	return 0;
}

