/*
 * 
 * Copyright (C) 2007:  Fabio Campisi (campisi@cli.di.unipi.it) 
 *                      Daniele Sgandurra (daniele@di.unipi.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * --------------------
 *
 * Author(s): Fabio Campisi (campisi@cli.di.unipi.it) 
 *            Daniele Sgandurra (daniele@di.unipi.it)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include "xen_vmi.h"

#define FREE(ptr) if(ptr) free(ptr);

#define LIST_SIZE 3
struct digest_info **digest_list;

int init_vmi(uint32_t dom)
{

     int i=0;
     unsigned char **msg;
     unsigned char *md_value;
     unsigned int md_len;

     uint32_t *sys_table;
     int sys_num;

     unsigned char **kernel_pages;
     int kernel_page_num;

     unsigned char **idt_table;

     struct digest_info *elem = NULL;

     digest_list = (struct digest_info **)(malloc(LIST_SIZE * sizeof(struct digest_info *)));

     /*get the digest of system call table*/
     get_sys_call_table(&sys_table, &sys_num, dom);
     msg = (unsigned char **)(malloc(sizeof(unsigned char *)));
     msg[0] = (unsigned char*)sys_table;
     calculate_hash(msg, sizeof(uint32_t) * sys_num, 1, &md_value, &md_len);
     elem = malloc(sizeof(struct digest_info));
     sprintf(elem->name, "sys_call_table");
     memcpy(elem->digest, md_value, DIGEST_SIZE);
     (digest_list)[0] = elem;
     FREE(sys_table);
     FREE(msg);
     FREE(md_value);

     /*get the digest of kernel text area*/
     get_kernel_text(&kernel_pages, &kernel_page_num, dom);
     calculate_hash(kernel_pages, XC_PAGE_SIZE, kernel_page_num, &md_value, &md_len);
     elem = malloc(sizeof(struct digest_info));
     sprintf(elem->name, "kernel_text_area");
     memcpy(elem->digest, md_value, DIGEST_SIZE);
     (digest_list)[1] = elem;
     for(i=0; i< kernel_page_num; i++)
	  FREE(kernel_pages[i]);
     FREE(kernel_pages);
     FREE(md_value);

     /*get the digest of interrupt table*/
     get_idt_table(&idt_table, dom);
     calculate_hash(idt_table, sizeof(trap_info_t), IDT_ENTRIES, &md_value, &md_len);
     elem = malloc(sizeof(struct digest_info));
     sprintf(elem->name, "interrupt_table");
     memcpy(elem->digest, md_value, DIGEST_SIZE);
     (digest_list)[2] = elem;
     for(i=0; i<IDT_ENTRIES; i++)
	  FREE(idt_table[i]);
     FREE(idt_table);

     return 0;
}

void clean_data(int sig)
{
     int i = 0;

     for(i=0; i<LIST_SIZE; i++)
	  FREE(digest_list[i]);
     FREE(digest_list);
     exit(0);
}


int check_hash(char *check_name, unsigned char *check_digest)
{
     int i = 0;

     for(i=0; i<LIST_SIZE; i++)
     {
	  if(!strcmp(digest_list[i]->name,check_name))
	  {
	       if(!memcmp((digest_list)[i]->digest,check_digest,DIGEST_SIZE))
		    return 1;
	       else
		    return 0;
	  }
     }
     
     return -1;
}

int main(int argc, char **argv)
{
     uint32_t dom = 0;
     int opt, time = 60;
     int i=0;
     unsigned char **msg;
     unsigned char *md_value;
     unsigned int md_len;
     uint32_t *sys_table;
     int sys_num;
     unsigned char **kernel_pages;
     int kernel_page_num;
     unsigned char **idt_table;
     struct sigaction action;
     char tmp[DIGEST_SIZE];
     int syscall = 0;
     int kernel_text = 0;
     int idt = 0;
     int processes = 0;
     int modules = 0;

     if(argc < 2)
     {
	  printf("xen_check_dom -d dom [-p][-m][-s][-k][-i] [-t time]\n"
		 "-d dom \t\t domain id\n"
		 "-p \t\t list process list\n"
		 "-m \t\t list loaded modules\n"
		 "-s \t\t check syscall table\n"
		 "-k \t\t check kernel text area \n"
		 "-i \t\t check interrupt table \n"
		 "-t time \t time in-between two consecutive checks\n");
	  return 0;
     }

     if(argc > 2)
     {
	  while((opt = getopt(argc, argv, "d:t:pmski")) != -1)
	       switch(opt) {
	       case 'd': dom = atoi(optarg); 
		    break;
	       case 't': time = atoi(optarg); 
		    break;
	       case 'p' : processes = 1;
		    break;
	       case 'm' : modules = 1;
		    break;

	       case 's' : syscall = 1;
		    break;

	       case 'k' : kernel_text = 1;
		    break;

	       case 'i' : idt = 1;
		    break;

	       default  : printf("unknow options - Options: [-t][-l][-p][-m][-s][-k][-i]\n");
		    return 0;
	       }
     }

     action.sa_handler=clean_data;
     sigaction(SIGINT,&action,NULL);
     sigaction(SIGTERM,&action,NULL);
     
     printf("\n");

     init_vmi(dom);
     while(1)
     {
	  printf("-------START---------\n");
	  if(syscall)
	  {
	       get_sys_call_table(&sys_table, &sys_num, dom);
	
	       msg = (unsigned char **)(malloc(sizeof(unsigned char *)));
	       msg[0] = (unsigned char*)sys_table;
	       calculate_hash(msg, sizeof(uint32_t) * sys_num, 1, &md_value, &md_len);
	       printf("System call table digest is:\n");
	       for(i=0; i<md_len; i++)
		    printf("%02x", md_value[i]);
	       sprintf(tmp, "sys_call_table");
	       if(check_hash(tmp, md_value))
		    printf("\tcorrect\n");
	       else
		    printf("\tNOT correct\n");
	       FREE(sys_table);
	       FREE(msg);
	       FREE(md_value);
	       
	  }

	  if(kernel_text)
	  {
	       get_kernel_text(&kernel_pages, &kernel_page_num, dom);
	       calculate_hash(kernel_pages, XC_PAGE_SIZE, kernel_page_num, &md_value, &md_len);
	       printf("Kernel text area digest is: \n");
	       for(i=0; i<md_len; i++)
		    printf("%02x", md_value[i]);
	       sprintf(tmp, "kernel_text_area");
	       if(check_hash(tmp,md_value))
		    printf("\tcorrect\n");
	       else
		    printf("\tNOT correct\n");	
	       for(i=0; i<kernel_page_num; i++)
		    FREE(kernel_pages[i]);
	       FREE(kernel_pages);
	       FREE(md_value);
	       
	  }

	  if(idt)
	  {
	       get_idt_table(&idt_table, dom);
	       calculate_hash(idt_table, sizeof(trap_info_t), IDT_ENTRIES, &md_value, &md_len);
	       printf("Idt table digest is: \n");
	       for(i=0; i<md_len; i++)
		    printf("%02x", md_value[i]);
	       sprintf(tmp, "interrupt_table");
	       if(check_hash(tmp,md_value))
		    printf("\tcorrect\n");
	       else printf("\tNOT correct\n");

	       for(i=0; i<IDT_ENTRIES; i++)
		    FREE(idt_table[i]);
	       FREE(idt_table);
	
	  }
	  if(modules)
	       get_module_list(dom);
	  if(processes)
	       get_process_list(dom);
	  print_promisc_if(dom);
	  printf("-------FINISH---------\n");
	  printf("\n");
	  sleep(time);
     }
     return 0;
}
