/*
 * 
 * Copyright (C) 2007:  Fabio Campisi (campisi@cli.di.unipi.it) 
 *                      Daniele Sgandurra (daniele@di.unipi.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * --------------------
 *
 * Author(s): Fabio Campisi (campisi@cli.di.unipi.it) 
 *            Daniele Sgandurra (daniele@di.unipi.it)
 *
 */

#include "xen_vmi_private.h"


void *get_mem_domu(int xc_handle, int cpu, int current_domid, void *guest_va, int perm, pgd_access pgd)
{
     void *tmp;
     unsigned long dim;
     void *mem;

     tmp = mapmem_domu(xc_handle, cpu, current_domid, guest_va, perm, pgd);
     if(tmp == NULL)
	  return NULL;
     dim = (XC_PAGE_SIZE-((unsigned long) tmp - page_addr(tmp)));
     mem = malloc(dim*sizeof(char));
     memcpy(mem, tmp, dim);
     munmap_page(tmp);
  
     return mem;

}



int get_kernel_text(unsigned char ***text_page, int *page_num, uint32_t dom)
{
     void *mem_domu;
     char *symbol_start = "_text";
     char *symbol_end = "_etext";
     uint32_t start_addr, end_addr, next_addr;
     int xc_handle = xc_interface_open();

     (*page_num)=0;
     start_addr = get_kernel_address_from_symbol(map_file, symbol_start);
     if(start_addr == -1)
     {
	  printf("Failed to retrieve %s\n", symbol_start);
	  return -1;
     } 
     end_addr = get_kernel_address_from_symbol(map_file, symbol_end);
     if(end_addr == -1)
     {
	  printf("Failed to retrieve %s\n", symbol_end);
	  return -1;
     }

     next_addr = start_addr;
     mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)next_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
     (*page_num)++;
     (*text_page) = (unsigned char **)(malloc(*page_num * sizeof(unsigned char *)));
     (*text_page)[*page_num-1] = (unsigned char *)(malloc(XC_PAGE_SIZE * sizeof(unsigned char )));
     memcpy((*text_page)[*page_num-1], mem_domu, XC_PAGE_SIZE);

     while((next_addr = next_addr + (XC_PAGE_SIZE)) < end_addr)
     {
	  (*page_num)++;
	  FREE(mem_domu);
	  mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)next_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
	  (*text_page)=realloc(*text_page, *page_num * sizeof(unsigned char *) );
	  (*text_page)[*page_num-1] = (unsigned char *)(malloc(XC_PAGE_SIZE * sizeof(unsigned char )));
	  memcpy((*text_page)[*page_num-1], mem_domu, XC_PAGE_SIZE);
     }
     FREE(mem_domu);
     xc_interface_close(xc_handle);
     return 0;
}

int get_sys_call_table(uint32_t **sys_table, int *sys_num, uint32_t dom)
{
     void *mem_domu;
     char *symbol = "sys_call_table";
     uint32_t *sys_call_table;
     int i=0;
     uint32_t symbol_addr;
     int xc_handle = xc_interface_open();

     symbol_addr = get_kernel_address_from_symbol(map_file, symbol);     
     mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)symbol_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);

     if(mem_domu == NULL)
     {
	  printf("Failed to retrieve %s\n", symbol);
	  return -1;
     }

     sys_call_table = (uint32_t *)(mem_domu);
     *sys_table = (uint32_t *)(malloc((NR_syscalls) * sizeof(uint32_t)));
     *sys_num = NR_syscalls-1;

     for(i=0; i< NR_syscalls; i++)
	  (*sys_table)[i] = (uint32_t)sys_call_table[i];
     FREE(mem_domu);

     xc_interface_close(xc_handle);
     return 0;
}


/*with paravirtualization "idt_table" is "trap_table"*/
int get_idt_table(unsigned char ***idt_table, uint32_t dom)
{
     char *symbol = "trap_table";
     trap_info_t *trap_table;
     int i = 0;
     void *mem_domu;
     uint32_t symbol_addr;
     int xc_handle = xc_interface_open();
     symbol_addr = get_kernel_address_from_symbol(map_file, symbol);
     mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)symbol_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);

     if(mem_domu == NULL)
     {
	  printf("Failed to retrieve %s\n", symbol);
	  return -1;
     }

     trap_table = (trap_info_t *)(mem_domu);
     (*idt_table) = (unsigned char **)malloc(sizeof (unsigned char *));
     (*idt_table)[i] = (unsigned char *)malloc(sizeof (trap_info_t ));
     memcpy((*idt_table)[i], trap_table, (sizeof (trap_info_t )));
     trap_table++;

     while(++i < IDT_ENTRIES)
     {
	  (*idt_table) = (unsigned char **)realloc((*idt_table),(i+1) * sizeof (unsigned char *));
	  (*idt_table)[i] = (unsigned char *)malloc(sizeof (trap_info_t ));
	  memcpy((*idt_table)[i], trap_table, (sizeof (trap_info_t )));
	  trap_table++;
     }
     FREE(mem_domu);

     xc_interface_close(xc_handle);
     return 0;
}


int print_promisc_if(uint32_t dom)
{
     uint32_t next_addr;
     struct net_device *dev;
     struct net_device **dev_p;
     char *symbol = "dev_base";
     void *mem_domu;
     uint32_t symbol_addr;
     int xc_handle = xc_interface_open();

     symbol_addr = get_kernel_address_from_symbol(map_file, symbol);    
     mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)symbol_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
     if(mem_domu == NULL)
     {
	  printf("Failed to retrieve %s\n", symbol);
	  return -1;
     }

     dev_p = (struct net_device **)(mem_domu);
     next_addr = (uint32_t)(*dev_p);
     printf("INTERFACES:\n");
     do{
	  FREE(mem_domu);
	  mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)next_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
	  dev = (struct net_device*)(mem_domu);
	  if(strstr(dev->name,"eth") != NULL)
	  {
	       if((dev->flags & IFF_PROMISC) || (dev->gflags && IFF_PROMISC) || dev->promiscuity)
		    printf("%s: PROMISC\n", dev->name);
	       else printf("%s: not promisc\n", dev->name);
	  }	
  
     }while((next_addr=(uint32_t)dev->next) != 0);
     
     FREE(mem_domu);
     xc_interface_close(xc_handle);
     return 0;
}

int get_process_list(uint32_t dom)
{
     struct task_struct *task;
     uint32_t init_addr, next_addr;
     char *symbol = "init_task";
     void *mem_domu;
     int xc_handle = xc_interface_open();

     init_addr = get_kernel_address_from_symbol(map_file, symbol);

     if(init_addr == -1)
     {
	  printf("Failed to retrieve %s\n", symbol);
	  return -1;
     }
     mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)init_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
     if(mem_domu == NULL)
     {
	  printf("Failed to retrieve %s\n", symbol);
	  return -1;
     }
     xc_domain_pause(xc_handle, dom);
     printf("PROCESS LIST:\n");
     printf("PID\tUID\tCOMMAND\n");
     while(next_addr != init_addr)
     {	
	  task = (struct task_struct *)mem_domu;
	  next_addr = (uint32_t)next_task(task);
	  if(task->pid) printf("%d\t%d\t%s\n", task->pid, task->uid, task->comm);
	  FREE(mem_domu);
	  mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)next_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
     }
     xc_domain_unpause(xc_handle, dom);

     FREE(mem_domu);
     xc_interface_close(xc_handle);
     return 0;
}

int get_module_list(uint32_t dom)
{
     uint32_t next_addr;
     struct module *mod;
     char *symbol = "modules";
     struct list_head *head_ptr;
     uint32_t next_next_mod, next_mod_addr;
     void *mem_domu;
     uint32_t symbol_addr;
     int xc_handle = xc_interface_open();

     xc_domain_pause(xc_handle, dom);
     symbol_addr = get_kernel_address_from_symbol(map_file, symbol);  
     mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)symbol_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
     
     if(mem_domu == NULL)
     {
	  printf("Failed to retrieve %s\n", symbol);
	  return -1;
     }
     head_ptr = (struct list_head *)(mem_domu);
     mod = list_entry(head_ptr, struct module, list);
     next_mod_addr = (uint32_t)next_module(mod);
     next_next_mod = (uint32_t)mod->list.next;
     printf("MODULE LIST:\n");
     printf("NAME\t\tSIZE\tLICENSE\n");
     while(symbol_addr != next_next_mod)
     {
	  FREE(mem_domu);
	  mem_domu = get_mem_domu(xc_handle, 0, dom, (void *)next_mod_addr, PROT_READ | PROT_WRITE, KERNEL_PGD);
	  if(mem_domu == NULL)
	  {
	       printf("Failed to retrieve %x\n", next_addr);
	       return -1;
	  }
	  mod = (struct module*)(mem_domu);
	  next_mod_addr = (uint32_t)next_module(mod);
	  next_next_mod = (uint32_t)mod->list.next;

	  if(strlen(mod->name) < 6) printf("[%s]\t\t",mod->name);
	  else printf("[%s]\t",mod->name);
	  printf("%lu\t", (mod->init_size + mod->core_size));
	  if(!mod->license_gplok) printf("NOT-GPL\n");
	  else printf("GPL\n");
	  
     }
     xc_domain_unpause(xc_handle, dom);

     FREE(mem_domu);

     return 0;
}
