/*
 * 
 * Copyright (C) 2007:  Fabio Campisi (campisi@cli.di.unipi.it) 
 *                      Daniele Sgandurra (daniele@di.unipi.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * --------------------
 *
 * Author(s): Fabio Campisi (campisi@cli.di.unipi.it) 
 *            Daniele Sgandurra (daniele@di.unipi.it)
 *
 */


#include <linux/module.h>
#include <linux/string.h>
#include <linux/list.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/netdevice.h>
#include <stdio.h>
#include <linux/file.h>
#include <linux/dcache.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/spinlock.h>
#include <linux/spinlock_types.h>


#ifdef linux_ver2616
typedef uint8_t xen_domain_handle_t[16];
typedef struct trap_info 
{
     uint8_t       vector;
     uint8_t       flags;
     uint16_t      cs;
     unsigned long address;
} trap_info_t;
typedef unsigned long long xen_pfn_t;
#define map_file "/boot/System.map-2.6.16-xenU"
#endif

#ifdef linux_ver2618
#define map_file "/boot/System.map-2.6.18-xenU"
#endif

#define MODULE_NAME_LEN (64 - sizeof(unsigned long))
#define SHA1_DIGEST_SIZE 20
#define IDT_ENTRIES 256
#define XC_PAGE_SHIFT 12
#define XC_PAGE_SIZE            (1UL << XC_PAGE_SHIFT)


#define next_module(m)	list_entry((m)->list.next, struct module, list)
#define prev_module(m)	list_entry((m)->list.prev, struct module, list)


typedef struct xc_dominfo
{
     uint32_t      domid;
     uint32_t      ssidref;
     unsigned int  dying:1, crashed:1, shutdown:1,
	  paused:1, blocked:1, running:1,
	  hvm:1;
     unsigned int  shutdown_reason; /* only meaningful if shutdown==1 */
     unsigned long nr_pages;
     unsigned long shared_info_frame;
     uint64_t      cpu_time;
     unsigned long max_memkb;
     unsigned int  nr_online_vcpus;
     unsigned int  max_vcpu_id;
     xen_domain_handle_t handle;
} xc_dominfo_t;

struct digest_info
{
     char name[MODULE_NAME_LEN];
     unsigned char digest[SHA1_DIGEST_SIZE];
};

extern int xc_domain_pause(int xc_handle, uint32_t domid);
extern int xc_domain_unpause(int xc_handle,  uint32_t domid);
extern int atoi (__const char *__nptr);
extern int munmap (void *__addr, size_t __len);
extern void *malloc(size_t size);
extern void *realloc(void *ptr, size_t size);
extern void free(void *ptr);
extern unsigned int sleep(unsigned int seconds);

/*per ora qui*/
#define MAX_LENGTH 1000
struct dentry_list
{
     char name[MAX_LENGTH];
     struct dentry_list *previous;
};

struct dentry_list *last_dentry = NULL;

#define PROT_READ	0x1		/* page can be read */
#define PROT_WRITE	0x2		/* page can be written */
typedef enum { KERNEL_PGD, PROCESS_PGD } pgd_access;
int xc_interface_open(void);
int xc_interface_close(int xc_handle);
#define page_addr(_a) (((unsigned long)(_a)) & 0xFFFFF000)
void * mapmem_domu(
    int xc_handle,
    int cpu,
    int current_domid,
    void *guest_va,
    int perm,
    pgd_access pgd);

unsigned long get_kernel_address_from_symbol(
	char *symbol_t,
	char *symbol);

#define munmap_page(ptr) (munmap((void *)page_addr((ptr)), XC_PAGE_SIZE))
#define FREE(ptr) if(ptr) free(ptr);
